/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2011. All Rights Reserved.
 *
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/
dojo.provide("jazz.ui.Select");

dojo.require("jazz.ui.ScrollableMenu");
dojo.require("jazz.ui.toolbar.MenuButton");

(function() {

/**
 * Constructs a new Select widget.
 * @name jazz.ui.Select
 * @class <p>A rich select widget for web UIs.</p>
 * 	<p>PROVISIONAL: This API is considered provisional and may be changed or removed in the
 * 	future with minimal or no warning.</p>
 * @constructor
 * @extends dijit._Widget
 * @extends dijit._Templated
 * @extends dijit._Container
 */
dojo.declare("jazz.ui.Select", [dijit._Widget, dijit._Templated, dijit._Container], {

	templatePath: dojo.moduleUrl("jazz.ui", "templates/Select.html"),

	/**
	 * The index of the selected item, or the item itself.
	 * @field
	 * @type Number|Widget
	 */
	selected: null,
	
	/**
	 * Callback which will be called with the newly selected value whenever the
	 * currently selected value changes.
	 * @field
	 * @type Function
	 */
	onChange: function() {},
	
	/**
	 * Text to show when nothing is selected.
	 * @field
	 * @type String
	 */
	defaultText: "",
	
	/**
	 * Specify whether the widget is enabled. If false, widget will be faded out, 
	 * no hover, and unclickable.
	 * @field
	 * @type Boolean
	 */
	enabled: true,


	/**
	 * Add an item to the select menu. Inherited from dijit._Container.
	 * @param {_Widget} widget The widget to add. Must be compatible with dijit menus (dijit.MenuItem or subclass).
	 * @param {Number} insertIndex The index to insert the item.
	 */	
	addChild: function(/*dijit._Widget*/ widget, /*int?*/ insertIndex){
		if (!this._menu) {
			//add it to containerNode as usual
			this.inherited(arguments);
		} else {
			//add it to the menu
			this.connect(widget, "onClick", dojo.hitch(this, this._wasClicked, widget));
			this._menu.addChild.apply(this._menu, arguments);
		}
	},
	
	startup: function() {
		this.set("selected", this.selected);
		this.inherited(arguments);
	},

	// internals
	
	_open: false,

	widgetsInTemplate: true,
	
	buildRendering: function() {
		this.inherited(arguments);
		if (this.selected == null) {
			var nodes = this.containerNode.childNodes, index = 0;
			for (var i = 0; i < nodes.length; i++) {
				if (nodes[i].nodeType == 3) continue;
				if (nodes[i].getAttribute("selected")) {
					this.selected = index;
					return;
				}
				index++;
			}
		}
		this.set("selected", null);
	},
	
	postCreate: function() {
		this._menuButton.focusNode.tabIndex = "-1";
		dojo.mixin(this._menuButton, {
			menuProvider: dojo.hitch(this, this._createMenu),
			around: this.domNode,
			focusNode: this.domNode
		});
		this.connect(this.domNode, "ondragstart", dojo.stopEvent);
		this.connect(this.domNode, "onselectstart", dojo.stopEvent);
	},

	_onClick: function(event) {
		if (!this.enabled) return;
		this._menuButton.open();
	},
	
	_onKeyPress: function(event) {
		if (!this.enabled) return;
		this._menuButton._handleKeypress(event);
	},

	_onMouseOver: function(event) {
		this._over = true;
		this._toggleHoverBorder(true);
	},
	
	_onMouseOut: function(event) {
		this._over = false;
		this._toggleHoverBorder(false);
	},

	_toggleHoverBorder: function(active) {
		dojo.toggleClass(this.domNode, 'active', this.enabled && (active || this._open || this._over));
	},
	
	_onMenuOpen: function() {
		this._open = true;
		this._toggleHoverBorder(true);
		this._menu.domNode.style.width = "";
		var menuWidth = dojo.position(this._menu.domNode).w;
		var thisWidth = dojo.position(this.domNode).w;
		if (menuWidth < thisWidth) {
			dojo.style(this._menu.domNode, 'width', thisWidth + 'px');
		}
		if (this.selected) {
			this._menu.setSelected(this.selected);
		}
	},

	_onMenuClose: function() {
		this._open = false;
		this._toggleHoverBorder(false);
	},

	_createMenu: function() {
		if (!this._menu) {
			this._menu = new jazz.ui.ScrollableMenu({
				selectable: true,
				onOpen: dojo.hitch(this, this._onMenuOpen),
				onClose: dojo.hitch(this, this._onMenuClose)
			});
			var widgets = dijit.findWidgets(this.containerNode);
			dojo.forEach(widgets, function(widget){
				this.connect(widget, "onClick", dojo.hitch(this, this._wasClicked, widget));
				this._menu.addChild(widget);
			}, this);
		}
		return this._menu;
	},

	_setSelectedAttr: function(item) {
		if (typeof item == "number") {
			var widgets = this.getChildren();
			if (item < widgets.length) {
				item = widgets[item];
			}
		}
		if (this.selected == item || (item && !item.isFocusable())) return;
		if (item != null && item.domNode) {
			dojo.create("tr", {"class": item.domNode.className, innerHTML: item.domNode.innerHTML}, this._selectLeft, "only");
			this._menu&&this._menu.setSelected(item);			
		} else {
			this._doDefaultText();
		}
		var oldtype = typeof this.selected;
		this.selected = item;
		if (oldtype !== "number") this.onChange(item);
	},
	
	_setDefaultTextAttr: function(text) {
		this.defaultText = text;
		if (!this.selected) this._doDefaultText();
	},
	
	_doDefaultText: function() {
		this._selectLeft.innerHTML = "<tbody><tr><td class=\"j-hint\">" + this.defaultText + "</td></tr></tbody>";
	},
	
	_setEnabledAttr: function(enabled) {
		this.enabled = enabled;
		dojo.toggleClass(this.domNode,"enabled",enabled);
		if (enabled)
			this.domNode.tabIndex = "0";
		else
			this.domNode.removeAttribute("tabIndex");
		this._menuButton.set('enabled',enabled);
	},

	_wasClicked: function(item) {
		this.set("selected", item);
	},
	
	// override getChildren to deal with our fake container magic
	getChildren: function(){
		return this._menu ? this._menu.getChildren() : dijit.findWidgets(this.containerNode);
	}, 

	destroy: function() {
		if (this._menu) this._menu.destroyRecursive();
		this.inherited(arguments);
	}
});
})();