/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2011. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/
(function() {
dojo.provide("jazz.util.binding");

dojo.require("jazz.util.dom");

/**
 * @namespace
 * <p>
 * Provides facilities to simplify reading and modifying XML documents by mapping DOM
 * nodes and sub-trees to simple JavaScript objects. The JavaScript object then provides a
 * simplified interface through which to read and/or manipulate the DOM, much like reading
 * JSON objects.
 * </p>
 * <p>
 * The binding interface is a JavaScript object that describes the mappings between the
 * XML structure and the object through which the client will interact with the DOM. For
 * read-only use cases, the <code>read()</code> function can be used and the interface
 * specifies properties that will be populated with the desired values. For read and write
 * use cases, the <code>bind()</code> method is used, and the interface specifies getter
 * and setter function names, where the functions will get and set the values on the DOM.
 * </p>
 * <p>
 * This is best illustrated through example. Consider the following XML snippet, which
 * represents a book with a title, number of pages, and one or more authors:
 * </p>
 * <pre>
 * &lt;book pages=&quot;822&quot; xmlns:dc=&quot;http://purl.org/dc/terms/&quot;&gt;
 *    &lt;dc:title&gt;Moby-Dick&lt;/dc:title&gt;
 *    &lt;authors&gt;
 *       &lt;dc:creator&gt;Herman Melville&lt;/dc:creator&gt;
 *    &lt;/authors&gt;
 * &lt;/book&gt;
 * </pre>
 * <p>
 * Once parsed into an XML DOM structure, the data can be read into a simple
 * JavaScript object by using the <code>read()</code> function and the following binding
 * interface:
 * </p>
 * <pre>
 * var Book = {
 *    namespaces: {
 *       dc: &quot;http://purl.org/dc/terms/&quot;
 *    },
 *    authors: [&quot;string&quot;, &quot;authors/dc:creator&quot;, &quot;*&quot;],
 *    pages: [&quot;number&quot;, &quot;@pages&quot;, &quot;1&quot;],
 *    title: [&quot;string&quot;, &quot;dc:title&quot;, &quot;1&quot;]
 * };
 * </pre>
 * <p>
 * Here, the <code>Book</code> interface describes an object with properties that map to
 * various nodes in the XML, along with their multiplicity and type (full details are
 * explained below). The following snippet illustrates how to read the data associated
 * with the book:
 * </p>
 * <pre>
 * dojo.require(&quot;jazz.util.binding&quot;);
 * ...
 * var bookElem = ... [reference to the DOM Element];
 * var book = jazz.util.binding.read(bookElem, Book);
 * var title = book.title;
 * var pages = book.pages;
 * var authors = book.authors;
 * for (var i=0;i&lt;authors.length;i++) {
 *    var author = authors[i];
 *    // do something
 * }
 * </pre>
 * <p>
 * Alternatively, if modifications to the DOM are required, or reading all the properties up-front
 * is undesirable for performance reasons, the <code>bind()</code> function can be used. Unlike with
 * the <code>read()</code> function, which extracts the data and maintains no further relationship
 * with the DOM after the call, the <code>bind()</code> function establishes a link between getter
 * and setter functions on the binding object and the DOM. Properties starting with <code>get</code>,
 * <code>is</code>, and <code>has</code> become getters, and properties starting with <code>set</code>
 * become setters. Here is another version of the <code>Book</code> interface that can be used with
 * the <code>bind()</code> method:
 * </p>
 * <pre>
 * var Book = {
 *    namespaces: {
 *       dc: &quot;http://purl.org/dc/terms/&quot;
 *    },
 *    getAuthors: [&quot;string&quot;, &quot;authors/dc:creator&quot;, &quot;*&quot;],
 *    getPages: [&quot;number&quot;, &quot;@pages&quot;, &quot;1&quot;],
 *    getTitle: [&quot;string&quot;, &quot;dc:title&quot;, &quot;1&quot;],
 *    setAuthors: [&quot;string&quot;, &quot;authors/dc:creator&quot;, &quot;*&quot;],
 *    setPages: [&quot;number&quot;, &quot;@pages&quot;, &quot;1&quot;],
 *    setTitle: [&quot;string&quot;, &quot;dc:title&quot;, &quot;1&quot;]
 * };
 * </pre>
 * <p>
 * And the following snippet illustrates how to read the book information using <code>bind()</code>:
 * </p>
 * <pre>
 * dojo.require(&quot;jazz.util.binding&quot;);
 * ...
 * var bookElem = ... [reference to the DOM Element];
 * var book = jazz.util.binding.bind(bookElem, Book);
 * var title = book.getTitle();
 * var pages = book.getPages();
 * var authors = book.getAuthors();
 * for (var i=0;i&lt;authors.length;i++) {
 *    var author = authors[i];
 *    // do something
 * }
 * </pre>
 * <p>
 * Similarly, you can call the setters to set new information, and this will modify the DOM
 * directly:
 * </p>
 * <pre>
 * book.setTitle(&quot;Oliver Twist&quot;);
 * book.setPages(480);
 * book.setAuthors([&quot;Charles Dickens&quot;]);
 * </pre>
 * <p>
 * These statements modify the DOM element producing the following result:
 * </p>
 * <pre>
 * &lt;book pages=&quot;480&quot; xmlns:dc=&quot;http://purl.org/dc/terms/&quot;&gt;
 *    &lt;dc:title&gt;Oliver Twist&lt;/dc:title&gt;
 *    &lt;authors&gt;
 *       &lt;dc:creator&gt;Charles Dickens&lt;/dc:creator&gt;
 *    &lt;/authors&gt;
 * &lt;/book&gt;
 * </pre>
 * 
 * <h2>Binding Interface</h2>
 * <p>
 * This section describes the various elements of a binding interface.
 * </p>
 * 
 * <h4>Namespaces</h4>
 * <p>
 * If any of the properties, getters, or setters have <a href="http://www.w3.org/TR/xpath/">XPaths</a>
 * that use XML namespace prefixes, the binding interface must have a <code>namespaces</code> property,
 * which is an object where each of its properties is an XML namespace prefix used in the binding
 * interface, and the value is the namespace URI, as a string. The namespace prefixes do not need
 * to match those in the DOM.
 * </p>
 * <p>
 * For example, if the <code>dc</code> prefix is used to refer to the
 * <code>http://purl.org/dc/terms/</code> namespace URI, and the <code>xml</code> prefix
 * is used to refer to <code>http://www.w3.org/XML/1998/namespace</code>, you could specify the
 * namespaces as:
 * </p>
 * <pre>
 * namespaces: {
 *    dc: &quot;http://purl.org/dc/terms/&quot;,
 *    xml: &quot;http://www.w3.org/XML/1998/namespace&quot;
 * }
 * </pre>
 * 
 * <h4>Properties/Getters/Setters</h4>
 * <p>
 * The remainder of the interface consists of either properties (for the <code>read()</code>
 * function), or getters and setters (for the <code>bind()</code> function), which take the
 * same form except for the name of the property. Each of these is specified as an array of
 * values that defines the type, multiplicity, and path to the DOM node it binds to.
 * </p>
 * <p>
 * The value is always an array of this form:
 * </p>
 * <p>
 * <code>[&lt;type&gt;, &lt;xpath&gt;, &lt;multiplicity&gt;]</code>
 * </p>
 * <p>
 * <b>type</b> - A string or object denoting the type of the value of the property, getter
 * return value, or setter parameter. Allowed values are <code>&quot;string&quot;</code>,
 * <code>&quot;boolean&quot;</code>, <code>&quot;number&quot;</code>, which correspond to the
 * primitive JavaScript types, and objects representing binding interfaces to allow nested
 * bindings. For example, if the <code>&lt;book&gt;</code> element were a child of a parent
 * <code>&lt;library&gt;</code> element, you could create a binding for the library with a
 * getter to retrieve all the books like this:
 * </p>
 * <p>
 * <code>getBooks: [Book, &quot;book&quot;, &quot;*&quot;]</code>
 * </p>
 * <p>
 * Calling <code>getBooks()</code> would return an array-like object of book objects that
 * implement the <code>Book</code> interface described above.
 * </p>
 * <p>
 * <b>xpath</b> - A simplified form of <a href="http://www.w3.org/TR/xpath/">XPath</a> that
 * refers to a DOM node to bind to, relative to the root element of the binding, i.e. the
 * element that was supplied in the <code>bind()</code> or <code>read()</code> function.
 * Only simple <a href="http://www.w3.org/TR/xpath/">XPaths</a> of this form are allowed:
 * </p>
 * <p>
 * <code>&lt;elem&gt;/&lt;elem&gt;/.../&lt;elemOrAttribute&gt</code>
 * </p>
 * <p>
 * Here are some examples:
 * </p>
 * <p>
 * <code>elem</code><br/>
 * <code>elem1/elem2/elem3</code><br/>
 * <code>@attribute</code><br/>
 * <code>elem1/elem2/@attribute</code><br/>
 * <code>a:elem1/b:elem2/@c:attribute</code><br/>
 * </p>
 * <p>
 * The evaluation of these <a href="http://www.w3.org/TR/xpath/">XPaths</a> is also done in
 * a simplified manner. Each segment is matched to a single node, the one that appears first,
 * except for the last segment which can match any number of sibling elements, i.e. elements
 * that share the same parent element.
 * </p>
 * <p>
 * <b>multiplicity</b> - Specifies whether or not the last segment of the
 * <a href="http://www.w3.org/TR/xpath/">XPath</a> should match a single node or all siblings
 * of a given parent element. The former is specified with the string &quot;<code>1</code>&quot;,
 * and the latter with the string &quot;<code>*</code>&quot;. Matching multiple attributes
 * is not allowed, and an <code>Error</code> will be thrown.
 * </p>
 * 
 * <h4>Accessors</h4>
 * <p>
 * An accessor is a getter or a property specified on a binding interface, which is used for
 * reading (accessing) data. The following behaviors apply to property values for the
 * <code>read()</code> function and return values for <code>bind()</code> getters.
 * </p>
 * <ul>
 * <li>Accessors evaluate to <code>undefined</code> if any of the elements or attributes in their
 *    <a href="http://www.w3.org/TR/xpath/">XPath</a> does not exist.</li>
 * <li>String accessors yield empty string if the element exists but contains no text nodes.</li>
 * <li>Boolean accessors always yield booleans if there is a value specified, specifically
 *    <code>true</code> if the text is exactly <code>&quot;true&quot;</code>, and
 *    <code>false</code> otherwise (e.g. <code>&quot;TrUe&quot;</code> will evaluate to
 *    <code>false</code>).
 *    </li>
 * <li>Number getters yield <code>NaN</code> if there was text available but could not be
 *    parsed into a number.</li>
 * </ul>
 * 
 * <h4>Multi-value Accessors</h4>
 * <p>
 * Accessors (properties or getters) that have a multiplicity of <code>&quot;*&quot;</code> produce
 * an array-like object containing all the matched values in the order they appear in the document.
 * There may be other unrelated nodes, such as other elements, in between them in the document,
 * but the accessor will only find the ones that match the <a href="http://www.w3.org/TR/xpath/">XPath</a>.
 * </p>
 * <p>
 * You must not assume that the object is an actual array, however it will have the values
 * accessible using array notation, e.g. <code>values[0]</code>, <code>values[1]</code>, etc.
 * There will be a <code>length</code> property indicating how many elements there are, as
 * well as <code>push</code> and <code>splice</code> methods that behave exactly as the array
 * methods do. Modifying this array-like object using <code>push</code> and
 * <code>splice</code> will modify the DOM directly, e.g. by adding or removing nodes. You
 * cannot modify an element by assigning a new value at a specific index; you should use
 * splice to replace a value. You should also use <code>push</code> and <code>splice</code>
 * methods for manipulating individual elements rather than calling the setter with a new
 * array, as this is much more efficient.
 * </p>
 * <p>
 * If there are no matching nodes for a multi-value accessor, an empty array-like object is
 * produced. Arguments to <code>splice</code> and <code>push</code> for object types (nested
 * bindings) do not need to be binding objects created using <code>bind</code>.
 * See the Setters section for more details.
 * </p>
 * 
 * <h4>Setters</h4>
 * <p>
 * Setters always expect a single argument which conforms to the type specified in the binding
 * interface, and modify the underlying DOM immediately to represent the new value. Setters
 * will create any elements or attributes that do not yet exist, and can be used to add new
 * elements and attributes.
 * </p>
 * <p>
 * For setters of type object (nested bindings), you do not need to pass in a binding object
 * created from <code>bind</code>. You can use any object that implements any subset
 * of the binding interface (specifically, the getters). All getters that are part of the
 * binding interface will be called, and the values will be inserted into new DOM nodes, as
 * if calling the equivalent setters.
 * </p>
 * <p>
 * For example, if you had a <code>&lt;library&gt;</code> parent element binding with the
 * following setter to assign books:
 * </p>
 * <p>
 * <code>setBooks: [Book, &quot;book&quot;, &quot;*&quot;]</code>
 * </p>
 * <p>
 * You could clear all the books and set new ones like this:
 * </p>
 * <pre>
 * myLibrary.setBooks([
 *    {
 *       getAuthors: function() {return [&quot;Jane Austen&quot;]},
 *       getTitle: function() {return &quot;Pride and Prejudice&quot;},
 *       getPages: function() {return 272}
 *    },
 *    {
 *       getAuthors: function() {return [&quot;William Shakespear&quot;]},
 *       getTitle: function() {return &quot;Macbeth&quot;},
 *       getPages: function() {return 96}
 *    }
 * ]);
 * </pre>
 * <p>
 * Alternatively, you can use an abbreviated format that, like with the <code>read()</code>
 * function, uses simple properties with values instead of getters that return values. Each
 * property is treated as if it were a getter prefixed with <code>&quot;get&quot;</code> that
 * returns the property&apos;s value. Here is the abbreviated form of the above snippet.
 * </p>
 * <p>
 * <pre>
 * myLibrary.setBooks([
 *    {
 *       authors: [&quot;Jane Austen&quot;],
 *       title: &quot;Pride and Prejudice&quot;,
 *       pages: 272
 *    },
 *    {
 *       authors: [&quot;William Shakespear&quot;],
 *       title: &quot;Macbeth&quot;,
 *       pages: 96
 *    }
 * ]);
 * </pre>
 * </p>
 * <p>
 * Simple non-binding objects and the abbreviated form above are also allowed in the
 * <code>push</code> and <code>splice</code> methods of the array-like objects returned by
 * multi-valued accessors.
 * </p>
 * <p>
 * In addition, the following behaviors apply to setters:
 * </p>
 * <ul>
 * <li>Setting a value to <code>null</code> or <code>undefined</code> removes the DOM node it
 *    binds to and its descendants.</li>
 * <li>String setters will throw <code>Error</code>s if you try to set anything other than a
 *    string, <code>null</code>, or <code>undefined</code>.</li>
 * <li>Boolean setters will throw <code>Error</code>s if you try to set anythong other than a
 *    boolean, <code>null</code>, or <code>undefined</code>.</li>
 * <li>Number setters will throw <code>Error</code>s if you try to set <code>NaN</code> or
 *    anything other than a number, <code>null</code>, or <code>undefined</code>.</li>
 * </ul>
 * 
 * <h4>Multi-value Setters</h4>
 * <p>
 * You can clear all existing values and set new values by using a setter with the multiplicity
 * <code>&quot;*&quot;</code>. The argument must be an array-like object with elements at
 * indices and a length property. As with single-value setters, the values for object types
 * (nested bindings) do not need to be binding implementations created from
 * <code>bind</code>, and can be in abbreviated form (see Setters section).
 * </p>
 * 
 * <h4>Limitations</h4>
 * <p>
 * The following limitations apply to all functions in this package. There may be additional
 * limitations for each individual function. These limitations may be removed in the future
 * without warning.
 * </p>
 * <ul>
 * <li>String value accessors may omit leading and trailing whitespace on some browsers.</li>
 * <li>Creating new attributes or elements with new namespaces (not yet otherwise
 *   defined in the document) is not supported.</li>
 * </ul>
 * 
 * @since 4.0
 */
jazz.util.binding = {
	
	/**
	 * <p>
	 * Creates and returns a JavaScript object that implements the given binding
	 * interface and operates on (is rooted at) the given XML DOM element.
	 * The binding interface consists of getter and setter function names, where
	 * properties starting with <code>get</code>, <code>is</code> and <code>has</code>
	 * become getters, and properties starting with <code>set</code> become setters
	 * (see package documentation for more details).
	 * </p>
	 * <p>
	 * Use this function to perform DOM modifications, or for reading where reading the
	 * complete binding up-front may adversely affect performance (e.g. if some properties
	 * are not always needed).
	 * </p>
	 * <p>
	 * The resulting binding object will have getter and setter functions matching the
	 * binding interface, and will also have an <code>elem</code> property which is the
	 * DOM Element that was supplied in this function. This also applies to nested bindings.
	 * </p>
	 * <h4>Restrictions</h4>
	 * <ul>
	 * <li>Binding interfaces must not change while any of its binding objects are being used.</li>
	 * <li>You must not edit the DOM independently between two binding calls in a way that can
	 *    interfere with the binding.</li>
	 * <li>You must not make any assumptions about when bindings are resolved into DOM nodes.
	 *    The <a href="http://www.w3.org/TR/xpath/">XPaths</a> may be evaluated lazily.
	 * </li>
	 * </ul>
	 * <h4>Limitations</h4>
	 * <p>
	 * You cannot use a binding object after its DOM Element is removed from its document. However,
	 * you can remove and add a binding object again at the same level of the same document, e.g.
	 * to reorder elements.
	 * </p>
	 * 
	 * @param {Element} elem The XML DOM Element to bind to.
	 * @param {object} bindInterface The binding interface to provide an implementation for
	 *    (see package description for details).
	 * @return {object} An implementation of the binding interface for the given element.
	 */
	bind: function(/* Element */ elem, /* object */ bindInterface) {
		var binding = {elem: elem};
		var found = false;
		for (var i in bindInterface) {
			if (found || i !== NAMESPACES) {
				// methods are initially stubs, and are lazily resolved when first one is called
				binding[i] = _createStub(elem, bindInterface, binding, i);
			}
			else {
				found = true;
			}
		}
		return binding;
	},
	
	/**
	 * <p>
	 * Reads the contents of the given XML DOM Element and its children, and extracts the
	 * requested values as defined in the binding interface. The interface here consists of
	 * property names (as opposed to getters and setters for the <code>bind()</code> function).
	 * The returned object is a simple JavaScript object containing those properties with
	 * values extracted from the DOM. This allows the reading of XML documents from the browser,
	 * with the ease of use of JSON.
	 * </p>
	 * <p>
	 * Use this function to read data from an XML DOM. This reads the data up-front, and the
	 * returned object has no subsequent link to the DOM as with the <code>bind()</code>
	 * function. As such, if some of the properties are not always needed, it may be more
	 * efficient to use <code>bind()</code> and selectively get only the required data.
	 * </p>
	 * 
	 * @param {Element} elem The XML DOM Element to read from.
	 * @param {object} bindInterface The binding interface describing the mapping between
	 *    properties and DOM nodes (see package description for details).
	 * @return {object} An object containing the values for the properties requested in the interface.
	 */
	read: function(/* Element */ elem, /* object */ bindInterface) {
		// create another binding interface with getters instead of properties
		var bindInterface2 = {
			namespaces: bindInterface.namespaces
		};
		var found = false;
		for (var i in bindInterface) {
			if (found || i !== NAMESPACES) {
				bindInterface2[GET + i] = bindInterface[i];
			}
			else {
				found = true;
			}
		}
		
		// resolve the getters
		var binding = {};
		_resolveBindings(elem, bindInterface2, binding);
		
		// call each getter and populate the result object
		var result = {};
		found = false;
		for (var i in bindInterface) {
			if (found || i !== NAMESPACES) {
				var value = binding[GET + i]();
				var type = bindInterface[i][0];
				var isObjType = (typeof(type) === OBJECT);
				if (bindInterface[i][2] === SINGLE) {
					if (isObjType && value) {
						value = _read(value.elem, type);
					}
				}
				else if (value) {
					// collection
					var array = [];
					var len = value.length;
					for (var j=0;j<len;++j) {
						array[j] = isObjType ? _read(value[j].elem, type) : value[j];
					}
					value = array;
				}
				result[i] = value;
			}
			else {
				found = true;
			}
		}
		return result;
	}
};

var dom = jazz.util.dom;
var _bind = jazz.util.binding.bind;
var _read = jazz.util.binding.read;

// constants
var TRUE = "true", FALSE = "false", SLASH = "/", BOOLEAN = "boolean", NUMBER = "number", STRING = "string",
	OBJECT = "object", UNDEFINED, FUNCTION = "function", COLON = ":", COMMA = ",", AT = 64, S = 115,
	SETTER_PREFIX = "set", NAMESPACES = "namespaces", ELEM_PROPERTY = "elem", SINGLE = "1", GET = "get";

var CAST_FUNC = {
	"boolean": function(str) {return str === TRUE},
	"number": function(str) {return parseFloat(str)},
	"string": function(str) {return str},
	"object": function(elem, bindInterface) {
		return _bind(elem, bindInterface);
	}
};
var CAST_FUNC_SET = {
	"boolean": function(bool) {
		var type = typeof(bool);
		if (type === BOOLEAN) {
			return bool ? TRUE : FALSE;
		}
		else if (bool !== null && bool !== UNDEFINED) {
			throw new Error("Invalid argument for binding; must be a boolean");
		}
	},
	"number": function(num) {
		var type = typeof(num);
		if (!isNaN(num) && type === NUMBER) {
			return num + "";
		}
		else if (num !== null && num !== UNDEFINED) {
			throw new Error("Invalid argument for binding; must be a number (not NaN)");
		}
	},
	"string": function(str) {
		var type = typeof(str);
		if (type === STRING || str === null || str === UNDEFINED) {
			return str;
		}
		throw new Error("Invalid argument for binding; must be a string");
	},
	"object": function(obj, bindInterface, elem) {
		var type = typeof(obj);
		if (type === OBJECT) {
			if (obj !== null) {
				// if the binding already had an elem, use it
				if (obj.elem) {
					return obj.elem;
				}
				// do work on-DOM to get correct namespace prefixes
				var newElem = dom.createElementNS(elem.ownerDocument, elem.namespaceURI || null, elem.nodeName);
				var parentNode = elem.parentNode;
				parentNode.appendChild(newElem);
				
				// create a partial binding interface with setters for all the getters in the object
				// that match getters on the binding interface
				var bindInterface2 = { namespaces: bindInterface.namespaces };
				for (var i in obj) {
					var descriptor = null;
					var value = obj[i];
					if (typeof(value) === FUNCTION) {
						// if it's a matching getter (i.e. not a setter)
						if (!_isSetter(i)) {
							descriptor = bindInterface[i];
						}
					}
					else {
						// shorthand notation; properties instead of getters
						var iLower = i.toLowerCase();
						for (var j in bindInterface) {
							if (_isSetter(j)) {
								// ends with i ignore case?
								if (i.length === j.length - 3 && j.toLowerCase().indexOf(iLower, 3) === 3) {
									descriptor = bindInterface[j];
									break;
								}
							}
						}
					}
					if (descriptor) {
						// make a setter for this getter
						bindInterface2[SETTER_PREFIX + i] = descriptor;
					}
				}
				
				// create the binding and call all the setters
				var newElemBinding = _bind(newElem, bindInterface2);
				for (var i in newElemBinding) {
					if (i !== ELEM_PROPERTY) {
						// get value, and pass to newElem setter
						var value = obj[i.substring(3)];
						if (typeof(value) === FUNCTION) {
							value = value.call(obj); 
						}
						newElemBinding[i].call(newElemBinding, value);
					}
				}
				
				// remove from DOM before returning
				parentNode.removeChild(newElem);
				return newElem;
			}
		}
		else if (obj !== UNDEFINED) {
			throw new Error("Invalid argument for binding; must be an object");
		}
	}
};

/*
 * Determine the appropriate function to get and set text under an element, and
 * to get the localName of a node.
 */
var _getLocalName = _getLocalNameW3C;
var _getText = _getTextW3C, _setText = _setTextW3C;
if (dojo.isIE) {
	if (dojo.isIE >= 9) {
		_getLocalName = _getLocalNameIE9AndUp;
		_getText = _getTextIE9AndUp;
		_setText = _setTextIE9AndUp;
	}
	else {
		_getLocalName = _getLocalNameIE8AndLower;
		_getText = _getTextIE8AndLower;
		_setText = _setTextIE8AndLower;
	}
}

/*
 * A lookup table for finding a getter function for a given node type, which returns a
 * normalized value (e.g. string or element).
 */
var GET_FUNC = [
	// 0: n/a
	null,
	// 1: ELEMENT
	function(elem) {
		return elem;
	},
	// 2: ATTRIBUTE
	function(elem, entry) {
		var node = dom.getAttributeNodeNS(elem, entry.nodeNamespaceURI, entry.nodeName);
		if (node) {
			return node.value;
		}
	},
	// 3: TEXT
	_getText
];

/*
 * A lookup table for finding a setter function for a given node type, which injects a
 * normalized value into the DOM.
 */
var SET_FUNC = [
	// 0: n/a
	null,
	// 1: ELEMENT
	function(elem, entry, bindingInfo, value) {
		var parentNode = elem.parentNode;
		if (value) {
			// swap elem for value (the new elem)
			parentNode.replaceChild(value, elem);
		}
		else {
			parentNode.removeChild(elem);
		}
		// unresolve bindings affected by this change
		_unresolveBindings(entry, bindingInfo);
	},
	// 2: ATTRIBUTE
	function(elem, entry, bindingInfo, value) {
		if (typeof(value) === STRING) {
			// setAttributeNS requires a qualified attribute name, so we need to look up the prefix
			var qName = entry.nodeName;
			if (entry.nodeNamespaceURI) {
				var prefix = dom.lookupPrefix(elem, entry.nodeNamespaceURI);
				if (prefix) {
					qName = prefix + COLON + qName;
				}
			}
			dom.setAttributeNS(elem, entry.nodeNamespaceURI, qName, value);
		}
		else {
			dom.removeAttributeNS(elem, entry.nodeNamespaceURI, entry.nodeName);
		}
	},
	// 3: TEXT
	function(elem, entry, bindingInfo, value) {
		if (typeof(value) === STRING) {
			_setText(elem, value);
		}
		else {
			// remove element & unresolve bindings affected by this removal
			elem.parentNode.removeChild(elem);
			_unresolveBindings(entry, bindingInfo);
		}
	}
];

function _createStub(/* Element */ elem, /* object */ bindInterface, /* object */ binding, /* string */ method) {
	// when the first stub is called, it will resolve each stub into an appropriate
	// binding method
	return function() {
		_resolveBindings(elem, bindInterface, binding);
		return binding[method].apply(binding, arguments);
	};
}

function _resolveBindings(/* Element */ elem, /* object */ bindInterface, /* object */ binding, /* boolean */ create) {
	// parse the xpaths into a combined tree-like structure for fast lookup
	var xpathTree = {};
	// combine bindings with same xpath into groups (partial binding interfaces)
	var methodGroups = {};
	// a map of arrays holding cached return values for entries. If there is a cached value, it's kept
	// at the first index of the array, otherwise the array length is 0.
	var caches = {};
	// a set of remaining method groups to bind, keyed by xpath, so that we can keep track of which
	// ones were left unbound at the end
	var remainingMethodGroups = {};
	// info related to the binding as a whole
	var bindingInfo = {
		binding: binding,
		bindInterface: bindInterface,
		elem: elem,
		methodGroups: methodGroups
	};
	var namespaces = bindInterface[NAMESPACES] || {};
	var found = false;
	for (var i in bindInterface) {
		if (found || i !== NAMESPACES) {
			var methodName = i;
			var descriptor = bindInterface[methodName];
			var xpath = descriptor[1];
			
			// methods to the same xpath are grouped together in a partial binding so that they can be
			// resolved again later as a group and coordinate between setting and getting
			var methodGroup = methodGroups[xpath];
			if (!methodGroup) {
				methodGroup = methodGroups[xpath] = remainingMethodGroups[xpath] = { namespaces: namespaces };
			}
			methodGroup[methodName] = descriptor;
			
			// entry (data object) for this binding method
			var type = descriptor[0];
			var entry = {
				descriptor: descriptor,
				methodName: methodName,
				type: type,
				nodeType: typeof(type) === STRING ? 3 : 1,
				xpath: xpath
			};
			var many = (descriptor[2] !== SINGLE);
			if (many) {
				entry.many = many;
			}
			
			// parse the xpath
			var localName, prefix = null, node = xpathTree, len = xpath.length;
			var idx = 0, idx2 = 0, idx3, isAttr = false, isLast = false;
			while (!isLast) {
				idx2 = xpath.indexOf(SLASH, idx);
				if (idx2 === -1) {
					idx2 = len;
					isLast = true;
					isAttr = (xpath.charCodeAt(idx) === AT);
					idx += isAttr ? 1 : 0;
				}
				idx3 = xpath.indexOf(COLON, idx);
				if (idx3 == -1 || idx3 >= idx2) {
					prefix = null;
					localName = xpath.substring(idx, idx2);
				}
				else {
					prefix = xpath.substring(idx, idx3);
					localName = xpath.substring(idx3 + 1, idx2);
				}
				idx = idx2 + 1;
				
				var namespaceURI = namespaces[prefix];
				if (!namespaceURI && prefix) {
					throw new Error("Undeclared namespace prefix \"" + prefix + "\" in binding interface");
				}
				if (isLast) {
					// attributes are only allowed on last segment of xpath
					if (isAttr) {
						entry.nodeType = 2; // attribute
						if (many) {
							throw new Error("Binding to multiple attributes is not allowed");
						}
					}
					if (isAttr || many) {
						entry.nodeName = localName;
						entry.nodeNamespaceURI = namespaceURI || null;
						break;
					}
				}
				var children = node.children;
				if (!children) {
					children = node.children = {};
					node.numChildren = 0;
				}
				var parent = node;
				node = null;
				var childEntries = children[localName];
				if (!childEntries) {
					childEntries = children[localName] = [];
				}
				else {
					var len2 = childEntries.length;
					for (var j=0;j<len2;++j) {
						var childEntry = childEntries[j];
						if (namespaceURI === childEntry.namespaceURI) {
							node = childEntry;
							break;
						}
					}
				}
				if (!node) {
					node = {};
					childEntries.push(node);
					++parent.numChildren;
					if (namespaceURI) {
						node.namespaceURI = namespaceURI;
					}
				}
			}
			var entries = node.entries;
			if (!entries) {
				entries = node.entries = [];
			}
			entries.push(entry);
		}
		else {
			found = true;
		}
	}
	
	// bind methods to existing nodes; create if requested
	_bindMethods(elem, bindingInfo, caches, xpathTree, remainingMethodGroups, create);
	
	// bind any remaining method groups that did not have elements to bind to
	for (var i in remainingMethodGroups) {
		var methodGroup = remainingMethodGroups[i];
		var found = false;
		for (var j in methodGroup) {
			if (found || j !== NAMESPACES) {
				var entry = methodGroup[j];
				var cache = _getEntryCache(entry, caches);
				binding[j] = _createUnresolvedBindingMethod(j, entry, cache, methodGroup, bindingInfo);
			}
			else {
				found = true;
			}
		}
	}
}

function _unresolveBindings(entry, bindingInfo) {
	// unresolve any binding methods affected by the removal of this entry's element subtree,
	// i.e. any method part of a method group whose xpath is prefixed by the xpath of the
	// current element being removed
	var xpath = entry.xpath;
	var len = xpath.length;
	var methodGroups = bindingInfo.methodGroups;
	var binding = bindingInfo.binding;
	for (var i in methodGroups) {
		// match whole xpath segments only
		if (i.indexOf(xpath) === 0 && (i.length === len || i.charAt(len) === SLASH)) {
			var methodGroup = methodGroups[i];
			for (var j in methodGroup) {
				if (j !== NAMESPACES) {
					// "unresolving" these bindings back to stubs will cause *all* bindings
					// to get re-resolved as soon as one of the stubs is called
					binding[j] = _createStub(bindingInfo.elem, bindingInfo.bindInterface, binding, j);
				}
			}
		}
	}
}

function _bindMethods(/* Element */ elem, /* object */ bindingInfo, /* object */ caches, /* object */ node, /* object */ remainingMethodGroups, /* boolean */ create) {
	// bind entries for this element
	var binding = bindingInfo.binding;
	var entries = node.entries;
	if (entries) {
		var len = entries.length;
		for (var i=0;i<len;++i) {
			var entry = entries[i];
			var cache = _getEntryCache(entry.descriptor, caches);
			binding[entry.methodName] = _createBindingMethod(elem, entry, cache, bindingInfo);
			delete remainingMethodGroups[entry.xpath];
		}
	}
	
	// recursively bind child elements
	var children = node.children;
	if (children) {
		var numChildren = node.numChildren;
		var child = elem.firstChild;
		while (child) {
			if (child.nodeType === 1) {
				var localName = child.localName || child.baseName;
				var childEntries = children[localName];
				if (childEntries) {
					var len = childEntries.length;
					for (var i=len-1;i>=0;--i) {
						var childEntry = childEntries[i];
						var namespaceURI = childEntry.namespaceURI;
						if ((namespaceURI && namespaceURI === child.namespaceURI) || namespaceURI == dom.lookupNamespaceURI(child, child.prefix)) {
							_bindMethods(child, bindingInfo, caches, childEntry, remainingMethodGroups, create);
							childEntries.splice(i, 1);
							--numChildren;
							if (--len === 0) {
								delete children[localName];
								break;
							}
						}
					}
					if (numChildren === 0) {
						break;
					}
				}
			}
			child = child.nextSibling;
		}
		
		// create the remaining ones, if requested
		if (create) {
			for (var i in children) {
				var childEntries = children[i];
				var len = childEntries.length;
				for (var j=0;j<len;++j) {
					var childEntry = childEntries[j];
					var prefix = dom.lookupPrefix(elem, childEntry.namespaceURI);
					var qName = prefix ? prefix + COLON + i : i;
					var newElem = dom.createElementNS(elem.ownerDocument, childEntry.namespaceURI || null, qName);
					elem.appendChild(newElem);
					_bindMethods(newElem, bindingInfo, caches, childEntry, remainingMethodGroups, true);
				}
			}
		}
	}
}

/*
 * Creates a binding method on an elem that already exists.
 */
function _createBindingMethod(/* Element */ elem, /* object */ entry, /* any[] */ cache, /* object */ bindingInfo) {
	// check whether it's a getter or setter method
	if (_isSetter(entry.methodName)) {
		return _createSetter(elem, entry, cache, bindingInfo);
	}
	return _createGetter(elem, entry, cache, bindingInfo);
}

/*
 * Creates a binding method on an elem that doesn't yet exist.
 */
function _createUnresolvedBindingMethod(/* string */ methodName, /* Array */ descriptor, /* any[] */ cache, /* object */ methodGroup, /* object */ bindingInfo) {
	// check whether it's a getter or setter method
	if (_isSetter(methodName)) {
		// return a setter method that creates the necessary element(s) then binds
		return function(value) {
			// only create and resolve if we're trying to set a value
			var binding = bindingInfo.binding;
			if (value !== null && value !== UNDEFINED) {
				// resolve this method group, creating elements as needed
				_resolveBindings(bindingInfo.elem, methodGroup, binding, true);
				
				// and fulfill the original call with the newly bound method
				return binding[methodName](value);
			}
		}
	}
	// return a getter method that returns undefined, because there's no dom node
	return function() {};
}

function _createGetter(/* Element */ elem, /* object */ entry, /* any[] */ cache, /* object */ bindingInfo) {
	if (entry.many) {
		// multi-value getter
		return function() {
			if (cache.length === 0) {
				// return an array-like object
				cache[0] = _createCollection(elem, entry, bindingInfo);
			}
			return cache[0];
		};
	}
	else {
		// single-value getter
		var type = entry.type;
		var castFunc = CAST_FUNC[typeof(type) === OBJECT ? OBJECT : type];
		var getFunc = GET_FUNC[entry.nodeType];
		return function() {
			if (cache.length === 0) {
				// chain: first get normalized value, then cast it
				var value = getFunc(elem, entry);
				var result;
				if (value !== null && value !== UNDEFINED) {
					// we found some content; cast it to appropriate javascript type
					result = castFunc(value, type);
				}
				cache[0] = result;
			}
			return cache[0];
		};
	}
}

function _createSetter(/* Element */ elem, /* object */ entry, /* any[] */ cache, /* object */ bindingInfo) {
	if (entry.many) {
		// multi-value setter
		return function(values) {
			// validate arg
			var type = typeof(values);
			if (values !== null && (type !== OBJECT || typeof(values.length) !== NUMBER) && values !== UNDEFINED) {
				throw new Error("Invalid argument to multi-valued setter; must be an array (but was " + type + ")");
			}
			// null/undefined equivalent to empty array
			values = values || [];
			// clear cache
			cache.length === 1 && cache.pop();
			// remove existing nodes first
			_forEachChild(elem, entry.nodeNamespaceURI, entry.nodeName, function(child) {
				elem.removeChild(child);
			});
			
			// we need to do some recursive calls to create each individual getter;
			// temporarily turn many off while we do this
			entry.many = false;
			// add new elements
			var len = values.length;
			for (var i=0;i<len;++i) {
				var prefix = dom.lookupPrefix(elem, entry.nodeNamespaceURI);
				var qName = prefix ? prefix + COLON + entry.nodeName : entry.nodeName;
				var newElem = dom.createElementNS(elem.ownerDocument, entry.nodeNamespaceURI || null, qName);
				elem.appendChild(newElem);
				// create a setter and call it right away
				_createSetter(newElem, entry, [], bindingInfo)(values[i]);
			}
			// done; turn many back on
			entry.many = true;
		};
	}
	else {
		// single-value setter
		var type = entry.type;
		var typeStr = typeof(type) === OBJECT ? OBJECT : type;
		var castFunc = CAST_FUNC_SET[typeStr];
		var castFuncGet = CAST_FUNC[typeStr];
		var setFunc = SET_FUNC[entry.nodeType];
		return function(value) {
			// chain: first cast to normalized value, then set
			var result;
			var castedValue = castFunc(value, type, elem);
			setFunc(elem, entry, bindingInfo, castedValue);
			if (castedValue !== null && castedValue !== UNDEFINED) {
				// cache and return normalized value back
				result = castFuncGet(castedValue, type);
			}
			return cache[0] = result;
		};
	}
}

function _createCollection(/* Element */ elem, /* object */ entry, /* object */ bindingInfo) {
	// return an array-like object
	var collection = {
		length: 0,
		push: function() {
			// arguments isn't a real array, can't concat it
			var args = [this.length, 0];
			var len = arguments.length;
			for (var i=0;i<len;++i) {
				args.push(arguments[i]);
			}
			this.splice.apply(this, args);
			return this.length;
		},
		splice: function(index, num) {
			// validate args
			if (index !== parseInt(index, 10) || num !== parseInt(num, 10) || num < 0) {
				throw new Error("Invalid argument to binding splice() first two arguments must be non-negative integers");
			}
			if (index < 0 || index + num > this.length) {
				throw new Error("Invalid argument to binding splice() index out of range (" + index + ")");
			}
			var i = 0;
			var _this = this;
			var sentinel;
			var removed = [];
			// remove elements and find the sentinel (elem we should insert before)
			_forEachChild(elem, entry.nodeNamespaceURI, entry.nodeName, function(child) {
				if (i >= index) {
					if (i < index + num) {
						elem.removeChild(child);
						removed.push(collection[i]);
					}
					else if (i == index + num) {
						sentinel = child;
					}
				}
				++i;
			});
			// fix indices & length
			var offset = arguments.length - 2 - num;
			if (offset != 0) {
				// shift elements over by offset
				if (offset > 0) {
					var start = this.length - 1, end = index + num - 1, step = -1;
				}
				else {
					var start = index + num, end = this.length, step = 1;
				}
				for (var i=start;i!==end;i+=step) {
					this[i+offset] = this[i];
				}
				// remove deleted indices
				for (var i=this.length+offset;i<this.length;++i) {
					delete this[i];
				}
				this.length += offset;
			}
			
			// add new elements
			var len = arguments.length - 2;
			for (var i=0;i<len;++i) {
				var value = arguments[i+2];
				if (value === null || value === UNDEFINED) {
					throw new Error("Invalid argument in binding splice(); null and undefined are not allowed");
				}
				var prefix = dom.lookupPrefix(elem, entry.nodeNamespaceURI);
				var qName = prefix ? prefix + COLON + entry.nodeName : entry.nodeName;
				var newElem = dom.createElementNS(elem.ownerDocument, entry.nodeNamespaceURI || null, qName);
				if (sentinel) {
					elem.insertBefore(newElem, sentinel);
				}
				else {
					elem.appendChild(newElem);
				}
				// create a single-valued setter and call it right away
				entry.many = false;
				// single-valued setters return the normalized set value
				this[index + i] = _createSetter(newElem, entry, [], bindingInfo)(value);
				entry.many = true;
			}
			return removed;
		}
	};
	// we need to do some recursive calls to create each individual getter;
	// temporarily turn many off while we do this
	entry.many = false;
	_forEachChild(elem, entry.nodeNamespaceURI, entry.nodeName, function(child) {
		// create a getter then call it right away, insert value into collection
		collection[collection.length++] = _createGetter(child, entry, [], bindingInfo)();
	});
	entry.many = true;
	return collection;
}

/*
 * Calls the given function for each child element of the given element with the given name and namespace URI.
 */
function _forEachChild(/* Element */ elem, /* string */ namespaceURI, /* string */ localName, /* function */ callback) {
	var child = elem.firstChild;
	while (child) {
		var next = child.nextSibling;
		if (child.nodeType === 1) {
			var childLocalName = child.localName || child.baseName;
			if (localName === childLocalName && ((namespaceURI && namespaceURI === child.namespaceURI) || namespaceURI == dom.lookupNamespaceURI(child, child.prefix))) {
				callback(child);
			}
		}
		child = next;
	}
}

/*
 * Returns an array in which to store the cached value for the given descriptor.
 * All getters and setters that have the same descriptor and differ only by method name
 * share the same cache.
 */
function _getEntryCache(/* Array */ descriptor, /* object */ caches) {
	var type = descriptor[0];
	var isObj = (typeof(type) === OBJECT);
	var key = (isObj ? OBJECT : type) + COMMA + descriptor[1] + COMMA + descriptor[2];
	if (isObj) {
		var items = caches[key] || (caches[key] = []);
		var len = items.length;
		for (var i=0;i<len;++i) {
			var item = items[i];
			if (item.type === type) {
				return item.value;
			}
		}
		var item = {type: type, value: []};
		items.push(item);
		return item.value;
	}
	else {
		return caches[key] || (caches[key] = []);
	}
}

/*
 * Get/set text and localName under an element for W3C-compliant browsers.
 */
function _getLocalNameW3C(/* Element */ elem) {
	return elem.localName;
}
function _getTextW3C(/* Element */ elem) {
	return elem.textContent;
}
function _setTextW3C(/* Element */ elem, /* string */ text) {
	elem.textContent = text;
}

/*
 * Get/set text and localName under an element for IE version 9 and later. Here the text
 * property depends on whether the DOM is MSXML (e.g. parsed from an XHR) or native DOM
 * (using DOMParser or dojox.xml.parser.parse()).
 */
function _getLocalNameIE9AndUp(/* Element */ elem) {
	return elem.localName || elem.baseName;
}
function _getTextIE9AndUp(/* Element */ elem) {
	return elem.textContent || elem.text || "";
}
function _setTextIE9AndUp(/* Element */ elem, /* string */ text) {
	if (elem instanceof Element) {
		// native DOM elements support W3C textContent property
		elem.textContent = text;
	}
	else {
		// MSXML
		elem.text = text;
	}
}

/*
 * Get/set text and localName under an element for IE version 8 and lower (MSXML).
 */
function _getLocalNameIE8AndLower(/* Element */ elem) {
	return elem.baseName;
}
function _getTextIE8AndLower(/* Element */ elem) {
	return elem.text;
}
function _setTextIE8AndLower(/* Element */ elem, /* string */ text) {
	elem.text = text;
}

/*
 * Returns whether or not the given method name is a setter (false means getter).
 */
function _isSetter(/* string */ name) {
	return name.charCodeAt(0) === S;
}
})();
