/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2012. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/
dojo.provide("jazz.app.i18n");

(function() {
	
dojo.require("dojo.date.locale");

jazz.app.i18n = {
	/**
	 * 	summary:
	 * 		Provides internationalization functions: 
	 * 			- getting and setting bidi preferences
	 * 			- getting and setting calendar type
	 * 			- getting and setting whether structured text is enabled
	 * 
	 * 		Exposes the following shims for accessing the proper date package:
 	 * 			- jazz.app.i18n.date: the date package, equivalent to dojo.date or dojox.date.buddhist
 	 * 	 		- jazz.app.i18n.datePackage: a string representing the current date package
 	 * 			- jazz.app.i18n.Date: the date object for the current locale, equivalent to built-in 
 	 *            JavaScript Date object or dojox.date.hebrew.Date
	 * 
	 *  usage notes:
	 *  	- access this object via 'static' methods (e.g. jazz.app.i18n.isBidiEnabled())
	 *  	- do not instantiate 
	 * 
	 */
	/* boolean */ isBidiEnabled: function() {
		// summary:
		//		returns the value of the "bidiEnabled" preference.
		if (!window.localStorage){
			return false;
		}
		var bidiOn = localStorage.getItem(_BIDI_ENABLED_PREFS);
		if (bidiOn && bidiOn == "true"){
			return true;
		}
		return false;
	},

	/* void */ setBidiEnabled: function(/* boolean */ isBidiEnabled) {
		// summary:
		//		sets into localStorage the item, which corresponds to "bidiEnabled" preferences. 
		if (!window.localStorage){
			return;
		}
		var oldItem = localStorage.getItem(_BIDI_ENABLED_PREFS);		
		if (oldItem != (isBidiEnabled? "true" : "false")){
			localStorage.setItem(_BIDI_ENABLED_PREFS, isBidiEnabled? "true" : "false");
			if (!isBidiEnabled){
				localStorage.removeItem(_BTD_PREFS);
			}
			dojo.publish(_BIDI_STORAGE_CHANGED,[(isBidiEnabled? "ltr":"none")]);
		}		 
	},
	
	/* string */ getBaseTextDir: function() {
		// summary:
		//		returns the value of the "baseTextDirection" preference.
		if (!window.localStorage){
			return false;
		}
		var btd = localStorage.getItem(_BTD_PREFS);
		if (!btd){
			return "ltr";
		}
		return btd;
	},

	/* void */ setBaseTextDir: function(/* string */ btd) {
		// summary:
		//		Sets into localStorage the item, which corresponds to "bidiEnabled" preferences. 
		if (!window.localStorage){
			return;
		}
		var oldItem = localStorage.getItem(_BTD_PREFS);
		if (oldItem != btd){
			localStorage.setItem(_BTD_PREFS, btd.toLowerCase());
			dojo.publish(_BIDI_STORAGE_CHANGED,[btd]);
		}
	},
	
	/* string */ getWidgetBaseTextDir: function() {
		// summary:
		//		returns the value of the "baseTextDirection" preference.
		//		or "none", if bidi isn't enabled.
		if (!window.localStorage){
			return null;
		}
		var bidiOn = localStorage.getItem(_BIDI_ENABLED_PREFS);
		if (!bidiOn || bidiOn != "true"){
			return "none";
		}		
		var btd = localStorage.getItem(_BTD_PREFS);
		if (!btd){
			return "ltr";
		}
		return btd;
	},
		
	/* string */ getCalendar: function() {
		// summary:
		//		returns the value of the "calendarType" preference.
		//		Possible values are: "gregorian", "buddhist", "hebrew", "islamic".
		if (window.localStorage){
			var clndr = localStorage.getItem(_CALTYPE_PREFS);
		}
		return clndr||"gregorian";
	},

	/* void */ setCalendar: function(/* string */ clndr) {
		// summary:
		//		Sets into localStorage the item, which corresponds to "calendarType" preferences. 
		if (!window.localStorage){
			return;
		}
		localStorage.setItem(_CALTYPE_PREFS, clndr.toLowerCase());
	},	
	
	/* string */ applySTSupport: function(str){
		var BidiConstants = {};
		// Regex for a Hebrew character
		BidiConstants.HEBREW_CHAR = new RegExp("[\\u05be\\u05c0\\u05c3\\u05c6\\u05d0-\\u05f4"
							+ BidiConstants.RLM + "\\ufb1d\\ufb1f-\\ufb28\\ufb2a-\\ufb4f]");  //$NON-NLS-1$ $NON-NLS-2$
		// Regex for an Arabic character
		BidiConstants.ARABIC_CHAR = new RegExp("[\\u061b\\u0621-\\u064a\\u066d-\\u066f\\u0671-\\u06d5"
        				    + "\\u06dd\\u06e5\\u06e6\\u06ee\\u06ef\\u06fa-\\u070d\\u0710\\u0712-\\u072f"
            				+ "\\u074d-\\u07a5\\u07b1\\ufb50-\\ufd3d\\ufd50-\\ufdfc\\ufe70-\\ufefc]"); //$NON-NLS-1$ $NON-NLS-2$ $NON-NLS-3$	
		var flag = true;
		var tempStr = str;
		for (var i = 0; i < tempStr.length; i++){
			c = tempStr.charAt(i);
			if ((c.match(BidiConstants.ARABIC_CHAR) || c.match(BidiConstants.HEBREW_CHAR)) == true)
			  	flag = true;
			else 
			   	flag = false;
		}
		// Last character is a Bidi character
		if (flag == false)
			tempStr = tempStr + "\u200E"; //$NON-NLS-1$
    	return tempStr;
     }
};

var prefs = "jazz.app.i18n.preferences.",
	_BIDI_ENABLED_PREFS = prefs + "bidiOn",
	_BTD_PREFS = prefs + "baseTextDirection",
	_CALTYPE_PREFS = prefs + "calType",
	_BIDI_STORAGE_CHANGED = "bidiStorageChanged";

// Date package shims
function _createDateShims() {
	var i18n = jazz.app.i18n, 
		calendar = i18n.getCalendar(),
		isGregorian = (calendar == "gregorian");
		
	if (!isGregorian) {
		dojo.requireIf(true, "dojox.date." + calendar);
		dojo.requireIf(true, "dojox.date." + calendar + ".locale");
	}
	
	i18n.date = isGregorian ? dojo.date : dojox.date[calendar];
	i18n.datePackage = isGregorian ? "" : "dojox.date." + calendar;
	i18n.Date = isGregorian ? Date : dojox.date[calendar].Date;
	
	if (!isGregorian) {
		var format = i18n.date.locale.format;
		i18n.date.locale.format = function(d,o) { return format(new dojox.date[calendar].Date(d),o); };
	}	
}

_createDateShims();

})();