/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  Use,
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
dojo.provide("jazz.ensemble.internal.catalog.CatalogService");

dojo.require("jazz.client.xhr");
dojo.require("jazz.util.binding");
dojo.require("dojo.DeferredList");

(function()
{
var binding			= jazz.util.binding;
var CatalogService	= jazz.ensemble.internal.catalog.CatalogService;

	/**
	 * load and consolidate always returns a dojo.deferred
	 * This will take either an array or a single uri 
	 * 
	 * This returns deferred. The callback will return the results of all the uris
	 * 
	 * {
	 *    results: [uri]=Error // null if no error
	 *  
	 *    catalog:
	 *    {
	 *    	entries:[0..n],
	 *      origins:[0..n],
	 *      categories:[0..n]
	 *    }
	 * }  
	 */
jazz.ensemble.internal.catalog.CatalogService.load = function(uris)
{
	// make it an array if it isn't already
	if (!dojo.isArray(uris)) uris = [uris];

	var results 		= [];
	var catalogs    	= [];
	var deferredList	= [];
	
	dojo.forEach(uris, function (uri){
		
		var deferred = jazz.client.xhrGet({handleAs: "xml", url: uri});
		
		deferred.addCallback(function(doc){
			results[uri]=null; // null means no errors
			catalogs.push(CatalogService.parseWidgetCatalog(doc, uri));
		});
	

		deferred.addErrback(function(error){
	   		results[uri]=error;
		});
		
		deferredList.push(deferred);
	});
	
	
	/**
	 * this is the deferred we return to our consumer
	 * we fire it after all of the list has returned
	 */
	var finalDeferred =  new dojo.Deferred();

	/**
	 * deferredList returns after all the lists are done
	 */
	(new dojo.DeferredList(deferredList)).addCallback(function(resultList){
		
		finalDeferred.callback({"results": results,
								"catalog": CatalogService.consolidateWidgetCatalogs(catalogs) 
								});
	}); 

	return finalDeferred;
}

jazz.ensemble.internal.catalog.CatalogService.consolidateWidgetCatalogs = function(catalogs)
{
	if (!dojo.isArray(catalogs)) catalogs = [catalogs];

	var entryHash	    = {};
	var categoryHash 	= {};
	var originHash		= {};

	dojo.forEach(catalogs, function(catalog){
	
		dojo.forEach(catalog.entries, 	function (entry)	{entryHash[entry.uri] 	= entry;  });
		dojo.forEach(catalog.categories,function (category)	{categoryHash[category] = true;	  });
		dojo.forEach(catalog.origins, 	function (origin)	{originHash[origin.uri] = origin; });
	
	});
	
	var entries = [], categories = [], origins=[];
	
	for (var uri 		in entryHash)	{entries.	push(entryHash[uri]);	}
	for (var category 	in categoryHash){categories.push(category);			}
	for (var uri 		in originHash)	{origins.	push(originHash[uri]);	}

	return 	{"entries"	:entries, "categories":categories, "origins"	:origins};
};

 
jazz.ensemble.internal.catalog.CatalogService.parseWidgetCatalog = function(doc, rootUri)
{
	var error = _getParseError(doc);
	if (error) throw new Error(error);

	var docElement = binding.read(doc.documentElement, CATALOG);
	
	var content = {origins:[], categories:[], entries:[]};

	var categoryHash = {} // a unique list of categories
	
	dojo.forEach(docElement.entries, function(entry){
		var id	= entry.id;
		var uri	= entry.uri;

		if (uri){
			entry.uri=uri;
		}
		else if (!uri && id){
			entry.uri=rootUri+"/"+id;
		}
		else{
			// TODO Remove this and fail silently 
			console.warn("ERROR: Catalog entry without URI or id:", entry.title);
			return;
		}
		dojo.forEach(entry.categories, function(c){
			categoryHash[c]=1;
		});
		
		entry.title 		= entry.title||"";
		entry.type 			= entry.type||"";
		entry.description 	= entry.description||"";
		entry.widgetUri 	= entry.widgetUri||"";
		entry.author 		= entry.author||"";
		entry.icon 			= entry.icon||"";
		entry.preview 		= entry.preview||"";
		entry.thumbnail 	= entry.thumbnail||"";
		entry.origin 		= entry.origin||"";
		entry.gadgetable 	= entry.gadgetable||false;

		content.entries.push(entry);
	});

	dojo.forEach(docElement.origins, function(o){
		content.origins.push({title:o.title, uri:o.uri});
	});

	for (var c in categoryHash){
		content.categories.push(c);	
	}
	categoryHash = null;

	return content;
}

var NAMESPACES = {
	dc: "http://purl.org/dc/terms/",
	ju: "http://jazz.net/ns/ui#",
	rdf: "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
	xml: "http://www.w3.org/XML/1998/namespace"
};

var PREFERENCE = {
	namespaces: NAMESPACES,
	name: ["string", "@ju:name", "1"],
	value: ["string", "@ju:value", "1"]
};

var ENTRY = {
	namespaces: NAMESPACES,
	author: ["string", "ju:author", "1"],
	categories: ["string", "ju:category", "*"],
	description: ["string", "dc:description", "1"],
	gadgetable: ["boolean", "ju:gadgetable", "1"],
	icon: ["string", "ju:icon/@rdf:resource", "1"],
	id: ["string", "@xml:id", "1"],
	origin: ["string", "ju:origin/@rdf:resource", "1"],
	preferences: [PREFERENCE, "ju:preference", "*"],
	preview: ["string", "ju:preview/@rdf:resource", "1"],
	thumbnail: ["string", "ju:thumbnail/@rdf:resource", "1"],
	title: ["string", "dc:title", "1"],
	type: ["string", "dc:type", "1"],
	uri: ["string", "@rdf:about", "1"],
	widgetUri: ["string", "ju:widgetUri", "1"]
};

var ORIGIN = {
	namespaces: NAMESPACES,
	title: ["string", "dc:title", "1"],
	uri: ["string", "@rdf:about", "1"]
};

var CATALOG = {
	namespaces: NAMESPACES,
	entries: [ENTRY, "ju:catalog-entry", "*"],
	origins: [ORIGIN, "ju:origin", "*"]
};

function _getParseError(document) {
	if (typeof(DOMParser) !== "undefined") { 
		if (document.documentElement && (document.documentElement.localName === 'parsererror')) {
			return document.documentElement.childNodes[0].nodeValue;
		}
	}
	else if (typeof(ActiveXObject) !== "undefined") {
		if (document.parseError && document.parseError.errorCode) {
			return document.parseError.reason;
		}
	}
	return null;
}

})();