/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2011. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/
dojo.provide("jazz.ui.toolbar.Button");

dojo.require("dijit._Templated");
dojo.require("dijit._Widget");
dojo.require("jazz.util.moduleResource");
dojo.require("jazz.util.html");

(function() {

var CLASS_NAME = "jazz-ui-toolbar-Button";

/**
 * Constructs a new toolbar button. All fields may be initialized in the widget constructor, or
 * can be changed after construction by using the set() function.
 * 
 * @name jazz.ui.toolbar.Button
 * @class <p>A toolbar button.</p>
 * 	<p>PROVISIONAL: This API is considered provisional and may be changed or removed in the
 * 	future with minimal or no warning.</p>
 * @constructor
 * @extends dijit._Widget
 * @extends dijit._Templated
 */ 
dojo.declare("jazz.ui.toolbar.Button", [dijit._Widget, dijit._Templated], /** @lends jazz.ui.toolbar.Button.prototype */ {
	
	/**
	 * Class to put on the button domNode.
	 * @type String
	 * @default ""
	 */
	"class": "",

	/**
	 * Specify whether the button is enabled.  If false, button will be faded out, 
	 * no hover, and unclickable.
	 * @type String
	 * @default true
	 */
	enabled: true,
	
	/**
	 * Href for the toolbar button. Navigates to the href url when clicked.
	 * @type String
	 * @default "#"
	 */
	href: "#",

	/**
	 * Class to add to the image dom node. Use this without providing imgSrc to provide an icon
	 * using a background image (useful for spriting).
	 * @type String
	 * @default ""
	 */
	imgClass: "",

	/**
	 * URL of a 16x16 pixel image to use as the button's icon.  The URL may be absolute, relative or 
	 * module-relative. The value is passed through jazz.util.moduleResource, so, for 
	 * example, jazz.ui/graphics/myimage.png will resolve properly.
	 * @type String
	 * @default ""
	 */
	imgSrc: "",
	
	/**
	 * If specified, a label will be shown beside the button's icon.
	 * @type String
	 * @default ""
	 */
	label: "",
	
	/**
	 * User defined function to handle clicks.
	 * @type Function
	 */
	onClick: function() {},
	
	/**
	 * The tooltip for the button. If this button has an icon, but no label, this serves as the alt text for the icon as well.
	 * @type String
	 * @default ""
	 */	
	title: "",
	
	/**
	 * Determines whether this widget is focusable or not.  Used by dijit._KeyNavContainer to determine keyboard focus.
	 * @noparams
	 */
	isFocusable: function(){
		return this.get("enabled") && this.focusNode && (dojo.style(this.domNode, "display") != "none");
	},
	
	/**
	 * Focuses the this button. Used by dijit._KeyNavContainer to set keyboard focus.
	 * @noparams
	 */
	focus: function(){
		dijit.focus(this.focusNode);
	},
	
	
	// Internals
	
	attributeMap: dojo.delegate(dijit._Widget.prototype.attributeMap, {
		tabIndex: ["focusNode", "titleNode"],
		href: "focusNode",
		title: "focusNode"
	}),
			
	templatePath : dojo.moduleUrl("jazz.ui", "toolbar/templates/Button.html"),

	_onClick: function(event) {
		if (!this.get("enabled")) {
			dojo.stopEvent(event);
			return;
		}
		if (this.href == "#") event.preventDefault();
		this.onClick(event);
	},
	_onKeyPress: function(e) {
		// handle spacebar
		var k = dojo.keys;
		if (e.charCode == k.SPACE) e.charOrCode = k.SPACE;
		switch (e.charOrCode) {
			case k.SPACE:
				this._onClick(e);
				if (this.href && this.href != "#") {
					// hack to make sure we check for saving before we navigate away
					var PlatformUI = dojo.getObject("net.jazz.ajax.ui.PlatformUI");
					if (!PlatformUI || !PlatformUI.getWorkbench().shouldVetoNavigation(this.href)) {
						location.href = this.href;
					}
				}
		}
	},
	_setVisibleAttr: function(visible) {
		this.domNode.style.display = visible ? "" : "none";
	},
	_setEnabledAttr: function(enabled){
		dojo.toggleClass(this.focusNode, CLASS_NAME+"-disabled", !enabled);
		dijit.setWaiState(this.focusNode, "disabled", !enabled);
		if (enabled) {
			this.focusNode.href = this.href;
		} else {
			this.focusNode.removeAttribute("href");
		}
		this.enabled = enabled;
	},
	_setImgClassAttr: function(imgClass) {
		if (!this._img || this._created) this._updateImage(this.imgSrc, imgClass);
		this.imgClass = imgClass;
	},
	_setImgSrcAttr: function(src){
		src = jazz.util.moduleResource(src);
		this._updateImage(src, this.imgClass);
		this.imgSrc = src;
	},
	_updateImage: function(src, imgClass) {
		if (src || imgClass) {
			if (!src && imgClass) src = this._blankGif;
			if (this._img) {
				this._img.src = src;
				this._img.className = "button-img" + (imgClass ? " " + imgClass : "");
				if (!this._img.parentNode || this._img.parentNode !== this.focusNode) {
					dojo.place(this._img, this.focusNode, "first");
				}
			} else {
				//if there's a label, don't show alt text (redundant info) - else, use the title attribute as alt text.
				this._img = dojo.create("img", {
					src: src,
					"class": "button-img" + (imgClass ? " " + imgClass : ""),
					alt: this.label ? "" : this.title
				}, this.focusNode, "first");
			}
		} else {
			this._img && this._img.parentNode && this._img.parentNode.removeChild(this._img);
		}
	},
	_setLabelAttr: function(label){
		this.label = label;
		this._updateLabel(label);
	},
	_updateLabel: function(label) {
		if (label) {
			if (this._img) this._img.alt = "";
			this.focusNode.title = this.title || "";
		} else {
			if (this._img) this._img.alt = this.title;
			this.focusNode.title = label || this.title;
		}	
		this._label.style.display = label ? "inline" : "none";
		this._label.innerHTML = jazz.util.html.escape(label);
		dojo.toggleClass(this.domNode, "has-icon", label && !!(this.imgSrc||this.imgClass));
	}
});

})();